# pylint: disable=too-many-lines
#
# File:   courseware/capa/inputtypes.py
#

"""
Module containing the problem elements which render into input objects

- textline
- textbox (aka codeinput)
- schematic
- choicegroup (aka radiogroup, checkboxgroup)
- imageinput  (for clickable image)
- optioninput (for option list)
- filesubmission (upload a file)
- crystallography
- vsepr_input
- drag_and_drop
- formulaequationinput
- chemicalequationinput

These are matched by *.html files templates/*.html which are django templates with the
actual html.

Each input type takes the xml tree as 'element', the previous answer as 'value', and the
graded status as'status'
"""

# TODO: make hints do something

# TODO: make all inputtypes actually render msg

# TODO: remove unused fields (e.g. 'hidden' in a few places)

# TODO: add validators so that content folks get better error messages.


# Possible todo: make inline the default for textlines and other "one-line" inputs.  It probably
# makes sense, but a bunch of problems have markup that assumes block.  Bigger TODO: figure out a
# general css and layout strategy for capa, document it, then implement it.


import json
import logging
import re
import shlex  # for splitting quoted strings
import sys
import time
from datetime import datetime

import html5lib
import nh3
import pyparsing
import six
from calc.preview import latex_preview
from chem import chemcalc
from lxml import etree

from xblocks_contrib.problem.capa.xqueue_interface import XQUEUE_TIMEOUT
from xblocks_contrib.problem.markup import HTML, Text
from xblocks_contrib.problem.stringify import stringify_children

from . import xqueue_interface
from .registry import TagRegistry
from .util import sanitize_html

log = logging.getLogger(__name__)

registry = TagRegistry()


class Status:
    """
    Problem status
    attributes: classname, display_name, display_tooltip
    """

    css_classes = {
        # status: css class
        "unsubmitted": "unanswered",
        "incomplete": "incorrect",
        "queued": "processing",
    }
    __slots__ = ("classname", "_status", "display_name", "display_tooltip")

    def __init__(self, status, gettext_func=str):
        self.classname = self.css_classes.get(status, status)
        _ = gettext_func
        names = {
            "correct": _("correct"),
            "incorrect": _("incorrect"),
            "partially-correct": _("partially correct"),
            "incomplete": _("incomplete"),
            "unanswered": _("unanswered"),
            "unsubmitted": _("unanswered"),
            "submitted": _("submitted"),
            "queued": _("processing"),
        }
        tooltips = {
            # Translators: these are tooltips that indicate the state of an assessment question
            "correct": _("This answer is correct."),
            "incorrect": _("This answer is incorrect."),
            "partially-correct": _("This answer is partially correct."),
            "queued": _("This answer is being processed."),
        }
        tooltips.update(dict.fromkeys(["incomplete", "unanswered", "unsubmitted"], _("Not yet answered.")))
        self.display_name = names.get(status, str(status))
        self.display_tooltip = tooltips.get(status, "")
        self._status = status or ""

    def __str__(self):
        return self._status

    def __repr__(self):
        return f"Status({self._status!r})"

    def __eq__(self, other):
        return self._status == str(other)

    def __hash__(self):
        return hash(str(self))


class Attribute:
    """
    Allows specifying required and optional attributes for input types.
    """

    # want to allow default to be None, but also allow required objects
    _sentinel = object()

    def __init__(  # pylint: disable=too-many-positional-arguments,too-many-arguments
        self, name, default=_sentinel, transform=None, validate=None, render=True
    ):
        """
        Define an attribute

        name (str): then name of the attribute--should be alphanumeric (valid for an XML attribute)

        default (any type): If not specified, this attribute is required.  If specified, use this as the default value
                        if the attribute is not specified.  Note that this value will not be transformed or validated.

        transform (function str -> any type): If not None, will be called to transform the parsed value into an internal
                        representation.

        validate (function str-or-return-type-of-tranform -> unit or exception): If not None, called to validate the
                       (possibly transformed) value of the attribute.  Should raise ValueError with a helpful message if
                       the value is invalid.

        render (bool): if False, don't include this attribute in the template context.
        """
        self.name = name
        self.default = default
        self.validate = validate
        self.transform = transform
        self.render = render

    def parse_from_xml(self, element):
        """
        Given an etree xml element that should have this attribute, do the obvious thing:
          - look for it.  raise ValueError if not found and required.
          - transform and validate.  pass through any exceptions from transform or validate.
        """
        val = element.get(self.name)
        if self.default == self._sentinel and val is None:
            raise ValueError(f"Missing required attribute {self.name}.")

        if val is None:
            # not required, so return default
            return self.default

        if self.transform is not None:
            val = self.transform(val)

        if self.validate is not None:
            self.validate(val)

        return val


class InputTypeBase:
    """
    Abstract base class for input types.
    """

    template = None

    def __init__(self, system, xml, state):
        """
        Instantiate an InputType class.  Arguments:

        - system    : LoncapaModule instance which provides OS, rendering, and user context.
                      Specifically, must have a render_template function.
        - xml       : Element tree of this Input element
        - state     : a dictionary with optional keys:
                      * 'value'  -- the current value of this input
                                    (what the student entered last time)
                      * 'id' -- the id of this input, typically
                                "{problem-location}_{response-num}_{input-num}"
                      * 'status' (submitted, unanswered, unsubmitted)
                      * 'input_state' -- dictionary containing any inputtype-specific state
                                        that has been preserved
                      * 'feedback' (dictionary containing keys for hints, errors, or other
                         feedback from previous attempt.  Specifically 'message', 'hint',
                         'hintmode'.  If 'hintmode' is 'always', the hint is always displayed.)
        """

        self.xml = xml
        self.tag = xml.tag
        self.capa_system = system

        # NOTE: ID should only come from one place.  If it comes from multiple,
        # we use state first, XML second (in case the xml changed, but we have
        # existing state with an old id). Since we don't make this guarantee,
        # we can swap this around in the future if there's a more logical
        # order.

        self.input_id = state.get("id", xml.get("id"))
        if self.input_id is None:
            raise ValueError(f"input id state is None. xml is {etree.tostring(xml)}")

        self.value = state.get("value", "")

        feedback = state.get("feedback", {})
        self.msg = feedback.get("message", "")
        self.hint = feedback.get("hint", "")
        self.hintmode = feedback.get("hintmode", None)
        self.input_state = state.get("input_state", {})
        self.answervariable = state.get("answervariable", None)
        self.response_data = state.get("response_data")

        # put hint above msg if it should be displayed
        if self.hintmode == "always":
            self.msg = HTML("{hint}<br/>{msg}" if self.msg else "{hint}").format(
                hint=HTML(self.hint), msg=HTML(self.msg)
            )

        self.status = state.get("status", "unanswered")

        try:
            # Pre-parse and process all the declared requirements.
            self.process_requirements()

            # Call subclass "constructor" -- means they don't have to worry about calling
            # super().__init__, and are isolated from changes to the input
            # constructor interface.
            self.setup()
        except Exception as err:  # pylint: disable=broad-exception-caught
            # Something went wrong: add xml to message, but keep the traceback
            msg = f"Error in xml '{etree.tostring(xml)}': {err} "
            six.reraise(Exception, Exception(msg), sys.exc_info()[2])

    @classmethod
    def get_attributes(cls):
        """
        Should return a list of Attribute objects (see docstring there for details). Subclasses should override.  e.g.

        return [Attribute('unicorn', True), Attribute('num_dragons', 12, transform=int), ...]
        """
        return []

    def process_requirements(self):
        """
        Subclasses can declare lists of required and optional attributes.  This
        function parses the input xml and pulls out those attributes.  This
        isolates most simple input types from needing to deal with xml parsing at all.

        Processes attributes, putting the results in the self.loaded_attributes dictionary.  Also creates a set
        self.to_render, containing the names of attributes that should be included in the context by default.
        """
        # Use local dicts and sets so that if there are exceptions, we don't
        # end up in a partially-initialized state.
        loaded = {}
        to_render = set()
        for attribute in self.get_attributes():
            loaded[attribute.name] = attribute.parse_from_xml(self.xml)
            if attribute.render:
                to_render.add(attribute.name)

        self.loaded_attributes = loaded
        self.to_render = to_render

    def setup(self):
        """
        InputTypes should override this to do any needed initialization.  It is called after the
        constructor, so all base attributes will be set.

        If this method raises an exception, it will be wrapped with a message that includes the
        problem xml.
        """

    def handle_ajax(self, dispatch, data):
        """
        InputTypes that need to handle specialized AJAX should override this.

        Input:
            dispatch: a string that can be used to determine how to handle the data passed in
            data: a dictionary containing the data that was sent with the ajax call

        Output:
            a dictionary object that can be serialized into JSON. This will be sent back to the Javascript.
        """

    def _get_render_context(self):
        """
        Should return a dictionary of keys needed to render the template for the input type.

        (Separate from get_html to faciliate testing of logic separately from the rendering)

        The default implementation gets the following rendering context: basic things like value, id, status, and msg,
        as well as everything in self.loaded_attributes, and everything returned by self._extra_context().

        This means that input types that only parse attributes and pass them to the template get everything they need,
        and don't need to override this method.
        """
        context = {
            "id": self.input_id,
            "value": self.value,
            "status": Status(self.status, self.capa_system.i18n.gettext),
            "msg": self.msg,
            "response_data": self.response_data,
            "STATIC_URL": self.capa_system.STATIC_URL,
            "describedby_html": HTML(""),
        }

        # Generate the list of ids to be used with the aria-describedby field.
        descriptions = []

        # If there is trailing text, add the id as the first element to the list before adding the status id
        if "trailing_text" in self.loaded_attributes and self.loaded_attributes["trailing_text"]:
            trailing_text_id = "trailing_text_" + self.input_id
            descriptions.append(trailing_text_id)

        # Every list should contain the status id
        status_id = "status_" + self.input_id
        descriptions.append(status_id)
        descriptions.extend(list(self.response_data.get("descriptions", {}).keys()))
        description_ids = " ".join(descriptions)
        context.update({"describedby_html": HTML('aria-describedby="{}"').format(description_ids)})

        context.update((a, v) for (a, v) in six.iteritems(self.loaded_attributes) if a in self.to_render)
        context.update(self._extra_context())
        if self.answervariable:
            context.update({"answervariable": self.answervariable})
        return context

    def _extra_context(self):
        """
        Subclasses can override this to return extra context that should be passed to their templates for rendering.

        This is useful when the input type requires computing new template variables from the parsed attributes.
        """
        return {}

    def get_html(self):
        """
        Return the html for this input, as an etree element.
        """
        if self.template is None:
            raise NotImplementedError(f"no rendering template specified for class {self.__class__}")

        context = self._get_render_context()

        html = self.capa_system.render_template(self.template, context).strip()

        try:
            output = etree.XML(html)
        except etree.XMLSyntaxError as ex:
            # If `html` contains attrs with no values, like `controls` in <audio controls src='smth'/>,
            # XML parser will raise exception, so wee fallback to html5parser,
            #  which will set empty "" values for such attrs.
            try:
                output = html5lib.parseFragment(html, treebuilder="lxml", namespaceHTMLElements=False)[0]
            except IndexError as exc:
                raise ex from exc

        return output

    def get_user_visible_answer(self, internal_answer):
        """
        Given the internal representation of the answer provided by the user, return the representation of the answer
        as the user saw it.  Subclasses should override this method if and only if the internal represenation of the
        answer is different from the answer that is displayed to the user.
        """
        return internal_answer


# -----------------------------------------------------------------------------


@registry.register
class OptionInput(InputTypeBase):
    """
    Input type for selecting and Select option input type.

    Example:

    <optioninput options="('Up', 'Down')" correct="Up"/><text>The location of the sky</text>

    # TODO: allow ordering to be randomized
    """

    template = "optioninput.html"
    tags = ["optioninput"]

    @staticmethod
    def parse_options(options):
        """
        Given options string, convert it into an ordered list of (option_id, option_description) tuples, where
        id==description for now.  TODO: make it possible to specify different id and descriptions.
        """
        # convert single quotes inside option values to html encoded string
        options = re.sub(r"([a-zA-Z])('|\\')([a-zA-Z])", r"\1&#39;\3", options)
        options = re.sub(r"\\'", r"&#39;", options)  # replace already escaped single quotes
        # parse the set of possible options
        lexer = shlex.shlex(options[1:-1])

        lexer.quotes = "'"
        # Allow options to be separated by whitespace as well as commas
        lexer.whitespace = ", "

        # remove quotes
        # convert escaped single quotes (html encoded string) back to single quotes
        tokens = [x[1:-1].replace("&#39;", "'") for x in lexer]

        # make list of (option_id, option_description), with description=id
        return [(t, t) for t in tokens]

    @classmethod
    def get_attributes(cls):
        """
        Convert options to a convenient format.
        """
        return [Attribute("options", transform=cls.parse_options), Attribute("inline", False)]

    def _extra_context(self):
        """
        Return extra context.
        """
        _ = self.capa_system.i18n.gettext
        return {"default_option_text": _("Select an option")}


# -----------------------------------------------------------------------------


# TODO: consolidate choicegroup, radiogroup, checkboxgroup after discussion of
# desired semantics.


@registry.register
class ChoiceGroup(InputTypeBase):
    """
    Radio button or checkbox inputs: multiple choice or true/false.

    TODO: allow order of choices to be randomized, following lon-capa spec.
    Use the "location" attribute, e.g., random, top, bottom.

    Example::

        <choicegroup>
        <choice correct="false" name="foil1">
            <text>This is foil One.</text>
        </choice>
        <choice correct="false" name="foil2">
            <text>This is foil Two.</text>
        </choice>
        <choice correct="true" name="foil3">
            <text>This is foil Three.</text>
        </choice>
        </choicegroup>
    """

    template = "choicegroup.html"
    tags = ["choicegroup", "radiogroup", "checkboxgroup"]

    def setup(self):
        i18n = self.capa_system.i18n
        # suffix is '' or [] to change the way the input is handled in --as a scalar or vector
        # value.  (VS: would be nice to make this less hackish).
        if self.tag == "choicegroup":
            self.suffix = ""
            self.html_input_type = "radio"
        elif self.tag == "radiogroup":
            self.html_input_type = "radio"
            self.suffix = "[]"
        elif self.tag == "checkboxgroup":
            self.html_input_type = "checkbox"
            self.suffix = "[]"
        else:
            _ = i18n.gettext
            # Translators: 'ChoiceGroup' is an input type and should not be translated.
            msg = _("ChoiceGroup: unexpected tag {tag_name}").format(tag_name=self.tag)
            raise Exception(msg)  # pylint: disable=broad-exception-raised

        self.choices = self.extract_choices(self.xml, i18n)
        self._choices_map = dict(
            self.choices,
        )

    @classmethod
    def get_attributes(cls):
        # Make '_' a no-op so we can scrape strings. Using function instead of
        #  `django.utils.translation.ugettext_noop` because Django cannot be imported in this file
        def _(text):
            return text

        return [Attribute("show_correctness", "always"), Attribute("submitted_message", _("Answer received."))]

    def _extra_context(self):
        return {"input_type": self.html_input_type, "choices": self.choices, "name_array_suffix": self.suffix}

    @staticmethod
    def extract_choices(element, i18n, text_only=False):
        """
        Extracts choices for a few input types, such as ChoiceGroup, RadioGroup and
        CheckboxGroup.

        returns list of (choice_name, choice_text) tuples

        By default it will return any XML tag in the choice (e.g. <choicehint>) unless text_only=True is passed.

        TODO: allow order of choices to be randomized, following lon-capa spec.  Use
        "location" attribute, ie random, top, bottom.
        """

        choices = []
        _ = i18n.gettext

        for choice in element:
            if choice.tag == "choice":
                if not text_only:
                    text = stringify_children(choice)
                else:
                    text = choice.text
                choices.append((choice.get("name"), text))
            else:
                if choice.tag != "compoundhint":
                    msg = Text("[capa.inputtypes.extract_choices] {error_message}").format(
                        error_message=Text(
                            # Translators: '<choice>' and '<compoundhint>' are tag names and should not be translated.
                            _("Expected a <choice> or <compoundhint> tag; got {given_tag} instead")
                        ).format(given_tag=choice.tag)
                    )
                    raise Exception(msg)  # pylint: disable=broad-exception-raised
        return choices

    def get_user_visible_answer(self, internal_answer):
        if isinstance(internal_answer, str):
            return self._choices_map[internal_answer]

        return [self._choices_map[i] for i in internal_answer]


# -----------------------------------------------------------------------------


@registry.register
class JSInput(InputTypeBase):
    """
    Inputtype for general javascript inputs. Intended to be used with
    customresponse.
    Loads in a sandboxed iframe to help prevent css and js conflicts between
    frame and top-level window.

    iframe sandbox whitelist:
        - allow-scripts
        - allow-popups
        - allow-forms
        - allow-pointer-lock
        - allow-downloads

    This in turn means that the iframe cannot directly access the top-level
    window elements.
    Example:

        <jsinput html_file="/static/test.html"
                 gradefn="grade"
                 height="500"
                 width="400"/>

    See the documentation in docs/data/source/course_data_formats/jsinput.rst
    for more information.
    """

    template = "jsinput.html"
    tags = ["jsinput"]

    @classmethod
    def get_attributes(cls):
        """
        Register the attributes.
        """
        return [
            Attribute("params", None),  # extra iframe params
            Attribute("html_file", None),
            Attribute("gradefn", "gradefn"),
            Attribute("get_statefn", None),  # Function to call in iframe
            #   to get current state.
            Attribute("initial_state", None),  # JSON string to be used as initial state
            Attribute("set_statefn", None),  # Function to call iframe to
            #   set state
            Attribute("width", "400"),  # iframe width
            Attribute("height", "300"),  # iframe height
            # Title for the iframe, which should be supplied by the author of the problem. Not translated
            # because we are in a class method and therefore do not have access to capa_system.i18n.
            # Note that the default "display name" for the problem is also not translated.
            Attribute("title", "Problem Remote Content"),
            # SOP will be relaxed only if this attribute is set to false.
            Attribute("sop", None),
        ]

    def _extra_context(self):
        context = {
            "jschannel_loader": f"{self.capa_system.STATIC_URL}js/capa/src/jschannel.js",
            "jsinput_loader": f"{self.capa_system.STATIC_URL}js/capa/src/jsinput.js",
            "saved_state": self.value,
        }

        return context


# -----------------------------------------------------------------------------


@registry.register
class TextLine(InputTypeBase):
    """
    A text line input.  Can do math preview if "math"="1" is specified.

    If "trailing_text" is set to a value, then the textline will be shown with
    the value after the text input, and before the checkmark or any input-specific
    feedback. HTML will not work, but properly escaped HTML characters will. This
    feature is useful if you would like to specify a specific type of units for the
    text input.

    If the hidden attribute is specified, the textline is hidden and the input id
    is stored in a div with name equal to the value of the hidden attribute.  This
    is used e.g. for embedding simulations turned into questions.

    Example:
        <textline math="1" trailing_text="m/s"/>

    This example will render out a text line with a math preview and the text 'm/s'
    after the end of the text line.
    """

    template = "textline.html"
    tags = ["textline"]

    @classmethod
    def get_attributes(cls):
        """
        Register the attributes.
        """
        return [
            Attribute("size", None),
            Attribute("hidden", False),
            Attribute("inline", False),
            # Attributes below used in setup(), not rendered directly.
            Attribute("math", None, render=False),
            # TODO: 'dojs' flag is temporary, for backwards compatibility with
            # 8.02x
            Attribute("dojs", None, render=False),
            Attribute("preprocessorClassName", None, render=False),
            Attribute("preprocessorSrc", None, render=False),
            Attribute("trailing_text", ""),
        ]

    def setup(self):
        self.do_math = bool(self.loaded_attributes["math"] or self.loaded_attributes["dojs"])

        # TODO: do math checking using ajax instead of using js, so
        # that we only have one math parser.
        self.preprocessor = None
        if self.do_math:
            # Preprocessor to insert between raw input and Mathjax
            self.preprocessor = {
                "class_name": self.loaded_attributes["preprocessorClassName"],
                "script_src": self.loaded_attributes["preprocessorSrc"],
            }
            if None in list(self.preprocessor.values()):
                self.preprocessor = None

    def _extra_context(self):
        return {
            "do_math": self.do_math,
            "preprocessor": self.preprocessor,
        }


# -----------------------------------------------------------------------------


@registry.register
class FileSubmission(InputTypeBase):
    """
    Upload some files (e.g. for programming assignments)
    """

    template = "filesubmission.html"
    tags = ["filesubmission"]

    @staticmethod
    def parse_files(files):
        """
        Given a string like 'a.py b.py c.out', split on whitespace and return as a json list.
        """
        return json.dumps(files.split())

    @classmethod
    def get_attributes(cls):
        """
        Convert the list of allowed files to a convenient format.
        """
        return [
            Attribute("allowed_files", "[]", transform=cls.parse_files),
            Attribute("required_files", "[]", transform=cls.parse_files),
        ]

    def setup(self):
        """
        Do some magic to handle queueing status (render as "queued" instead of "incomplete"),
        pull queue_len from the msg field.  (TODO: get rid of the queue_len hack).
        """
        _ = self.capa_system.i18n.gettext
        submitted_msg = _(
            "Your files have been submitted. As soon as your submission is"
            " graded, this message will be replaced with the grader's feedback."
        )
        self.submitted_msg = submitted_msg

        # Check if problem has been queued
        self.queue_len = 0
        # Flag indicating that the problem has been queued, 'msg' is length of
        # queue
        if self.status == "incomplete":
            self.status = "queued"
            self.queue_len = self.msg
            self.msg = self.submitted_msg

    def _extra_context(self):
        return {
            "queue_len": self.queue_len,
        }


# -----------------------------------------------------------------------------


@registry.register
class CodeInput(InputTypeBase):
    """
    A text area input for code--uses codemirror, does syntax highlighting, special tab handling,
    etc.
    """

    template = "codeinput.html"
    tags = [
        "codeinput",
        "textbox",
        # Another (older) name--at some point we may want to make it use a
        # non-codemirror editor.
    ]

    @classmethod
    def get_attributes(cls):
        """
        Convert options to a convenient format.
        """
        return [
            Attribute("rows", "30"),
            Attribute("cols", "80"),
            Attribute("hidden", ""),
            # For CodeMirror
            Attribute("mode", "python"),
            Attribute("linenumbers", "true"),
            # Template expects tabsize to be an int it can do math with
            Attribute("tabsize", 4, transform=int),
        ]

    def setup_code_response_rendering(self):
        """
        Implement special logic: handle queueing state, and default input.
        """
        # if no student input yet, then use the default input given by the
        # problem
        if not self.value and self.xml.text:
            self.value = self.xml.text.strip()

        # Check if problem has been queued
        self.queue_len = 0  # pylint: disable=attribute-defined-outside-init
        # Flag indicating that the problem has been queued, 'msg' is length of
        # queue
        if self.status == "incomplete":
            self.status = "queued"
            self.queue_len = self.msg  # pylint: disable=attribute-defined-outside-init
            self.msg = nh3.clean(self.submitted_msg)

    def setup(self):
        """setup this input type"""
        _ = self.capa_system.i18n.gettext
        submitted_msg = _(
            "Your answer has been submitted. As soon as your submission is"
            " graded, this message will be replaced with the grader's feedback."
        )
        self.submitted_msg = submitted_msg

        self.setup_code_response_rendering()

    def _extra_context(self):
        """
        Define queue_len, arial_label and code mirror exit message context variables
        """
        _ = self.capa_system.i18n.gettext
        return {
            "queue_len": self.queue_len,
            "aria_label": _("{programming_language} editor").format(
                programming_language=self.loaded_attributes.get("mode")
            ),
            "code_mirror_exit_message": _("Press ESC then TAB or click outside of the code editor to exit"),
        }


# -----------------------------------------------------------------------------


@registry.register
class MatlabInput(CodeInput):
    """
    InputType for handling Matlab code input

    Example::

        <matlabinput rows="10" cols="80" tabsize="4">
            Initial Text
        </matlabinput>

    """

    template = "matlabinput.html"
    tags = ["matlabinput"]

    def setup(self):
        """
        Handle matlab-specific parsing
        """
        _ = self.capa_system.i18n.gettext

        submitted_msg = _(
            "Submitted. As soon as a response is returned, this message will be replaced by that feedback."
        )
        self.submitted_msg = submitted_msg

        self.setup_code_response_rendering()

        xml = self.xml

        self.plot_payload = xml.findtext("./plot_payload")
        # Check if problem has been queued
        self.queuename = "matlab"
        self.queue_msg = ""
        # this is only set if we don't have a graded response
        # the graded response takes precedence
        if "queue_msg" in self.input_state and self.status in ["queued", "incomplete", "unsubmitted"]:
            self.queue_msg = sanitize_html(self.input_state["queue_msg"])

        if "queuestate" in self.input_state and self.input_state["queuestate"] == "queued":
            self.status = "queued"
            self.queue_len = 1
            self.msg = self.submitted_msg
            # Handle situation if no response from xqueue arrived during specified time.
            if "queuetime" not in self.input_state or time.time() - self.input_state["queuetime"] > XQUEUE_TIMEOUT:
                self.queue_len = 0
                self.status = "unsubmitted"
                self.msg = _("No response from Xqueue within {xqueue_timeout} seconds. Aborted.").format(
                    xqueue_timeout=XQUEUE_TIMEOUT
                )

    def handle_ajax(self, dispatch, data):
        """
        Handle AJAX calls directed to this input

        Args:
            - dispatch (str) - indicates how we want this ajax call to be handled
            - data (dict) - dictionary of key-value pairs that contain useful data
        Returns:
            dict - 'success' - whether or not we successfully queued this submission
                 - 'message' - message to be rendered in case of error
        """

        if dispatch == "plot":
            return self._plot_data(data)
        return {}

    def ungraded_response(self, queue_msg, queuekey):
        """
        Handle the response from the XQueue
        Stores the response in the input_state so it can be rendered later

        Args:
            - queue_msg (str) - message returned from the queue. The message to be rendered
            - queuekey (str) - a key passed to the queue. Will be matched up to verify that
                            this is the response we're waiting for

        Returns:
            nothing
        """
        # check the queuekey against the saved queuekey
        if (
            "queuestate" in self.input_state
            and self.input_state["queuestate"] == "queued"
            and self.input_state["queuekey"] == queuekey
        ):
            msg = self._parse_data(queue_msg)
            # save the queue message so that it can be rendered later
            self.input_state["queue_msg"] = msg
            self.input_state["queuestate"] = None
            self.input_state["queuekey"] = None

    def button_enabled(self):
        """Return whether or not we want the 'Test Code' button visible

        Right now, we only want this button to show up when a problem has not been
        checked.
        """
        if self.status in ["correct", "incorrect", "partially-correct"]:
            return False

        return True

    def _extra_context(self):
        """Set up additional context variables"""

        _ = self.capa_system.i18n.gettext

        queue_msg = self.queue_msg
        if len(self.queue_msg) > 0:  # An empty string cannot be parsed as XML but is okay to include in the template.
            try:
                etree.XML(HTML("<div>{0}</div>").format(HTML(self.queue_msg)))
            except etree.XMLSyntaxError:
                try:
                    html5lib.parseFragment(self.queue_msg, treebuilder="lxml", namespaceHTMLElements=False)[0]
                except (IndexError, ValueError):
                    # If neither can parse queue_msg, it contains invalid xml.
                    queue_msg = HTML("<span>{0}</span>").format(_("Error running code."))

        extra_context = {
            "queue_len": str(self.queue_len),
            "queue_msg": queue_msg,
            "button_enabled": self.button_enabled(),
            "matlab_editor_js": f"{self.capa_system.STATIC_URL}js/vendor/CodeMirror/octave.js",
            "msg": sanitize_html(self.msg),  # sanitize msg before rendering into template
        }
        return extra_context

    def _parse_data(self, queue_msg):
        """
        Parses the message out of the queue message
        Args:
            queue_msg (str) - a JSON encoded string
        Returns:
            returns the value for the the key 'msg' in queue_msg
        """
        try:
            result = json.loads(queue_msg)
        except (TypeError, ValueError):
            log.error("External message should be a JSON serialized dict. Received queue_msg = %s", queue_msg)
            raise
        msg = result["msg"]
        return msg

    def _plot_data(self, data):
        """
        AJAX handler for the plot button
        Args:
            get (dict) - should have key 'submission' which contains the student submission
        Returns:
            dict - 'success' - whether or not we successfully queued this submission
                 - 'message' - message to be rendered in case of error
        """
        _ = self.capa_system.i18n.gettext
        # only send data if xqueue exists
        if self.capa_system.xqueue is None:
            return {"success": False, "message": _("Cannot connect to the queue")}

        # pull relevant info out of get
        response = data["submission"]

        # construct xqueue headers
        qinterface = self.capa_system.xqueue.interface
        qtime = datetime.utcnow().strftime(xqueue_interface.DATEFORMAT)
        callback_url = self.capa_system.xqueue.construct_callback("ungraded_response")
        anonymous_student_id = self.capa_system.anonymous_student_id
        # TODO: Why is this using self.capa_system.seed when we have self.seed???
        queuekey = xqueue_interface.make_hashkey(
            str(self.capa_system.seed) + qtime + anonymous_student_id + self.input_id
        )
        xheader = xqueue_interface.make_xheader(
            lms_callback_url=callback_url, lms_key=queuekey, queue_name=self.queuename
        )

        # construct xqueue body
        student_info = {"anonymous_student_id": anonymous_student_id, "submission_time": qtime}
        contents = {
            "grader_payload": self.plot_payload,
            "student_info": json.dumps(student_info),
            "student_response": response,
            "token": getattr(self.capa_system, "matlab_api_key", None),
            "endpoint_version": "2",
            "requestor_id": anonymous_student_id,
        }

        error, msg = qinterface.send_to_queue(header=xheader, body=json.dumps(contents))
        # save the input state if successful
        if error == 0:
            self.input_state["queuekey"] = queuekey
            self.input_state["queuestate"] = "queued"
            self.input_state["queuetime"] = time.time()

        return {"success": error == 0, "message": msg}


# -----------------------------------------------------------------------------


@registry.register
class Schematic(InputTypeBase):
    """
    InputType for the schematic editor
    """

    template = "schematicinput.html"
    tags = ["schematic"]

    @classmethod
    def get_attributes(cls):
        """
        Convert options to a convenient format.
        """
        return [
            Attribute("height", None),
            Attribute("width", None),
            Attribute("parts", None),
            Attribute("analyses", None),
            Attribute("initial_value", None),
            Attribute("submit_analyses", None),
        ]

    def _extra_context(self):
        context = {
            "setup_script": f"{self.capa_system.STATIC_URL}js/capa/schematicinput.js",
        }

        return context


# -----------------------------------------------------------------------------


@registry.register
class ImageInput(InputTypeBase):
    """
    Clickable image as an input field.  Element should specify the image source, height,
    and width, e.g.

    <imageinput src="/static/Figures/Skier-conservation-of-energy.jpg" width="388" height="560" />

    TODO: showanswer for imageimput does not work yet - need javascript to put rectangle
    over acceptable area of image.
    """

    template = "imageinput.html"
    tags = ["imageinput"]

    @classmethod
    def get_attributes(cls):
        """
        Note: src, height, and width are all required.
        """
        return [
            Attribute("src"),
            Attribute("height"),
            Attribute("width"),
        ]

    def setup(self):
        """
        if value is of the form [x,y] then parse it and send along coordinates of previous answer
        """
        m = re.match(r"\[([0-9]+),([0-9]+)]", self.value.strip().replace(" ", ""))
        if m:
            # Note: we subtract 15 to compensate for the size of the dot on the screen.
            # (is a 30x30 image--lms/static/images/green-pointer.png).
            self.gx, self.gy = [int(x) - 15 for x in m.groups()]
        else:
            self.gx, self.gy = (0, 0)

    def _extra_context(self):

        return {"gx": self.gx, "gy": self.gy}


# -----------------------------------------------------------------------------


@registry.register
class Crystallography(InputTypeBase):
    """
    An input for crystallography -- user selects 3 points on the axes, and we get a plane.

    TODO: what's the actual value format?
    """

    template = "crystallography.html"
    tags = ["crystallography"]

    @classmethod
    def get_attributes(cls):
        """
        Note: height, width are required.
        """
        return [
            Attribute("height"),
            Attribute("width"),
        ]


# -------------------------------------------------------------------------


@registry.register
class VseprInput(InputTypeBase):
    """
    Input for molecular geometry--show possible structures, let student
    pick structure and label positions with atoms or electron pairs.
    """

    template = "vsepr_input.html"
    tags = ["vsepr_input"]

    @classmethod
    def get_attributes(cls):
        """
        Note: height, width, molecules and geometries are required.
        """
        return [
            Attribute("height"),
            Attribute("width"),
            Attribute("molecules"),
            Attribute("geometries"),
        ]


# -------------------------------------------------------------------------


@registry.register
class ChemicalEquationInput(InputTypeBase):
    """
    An input type for entering chemical equations.  Supports live preview.

    Example:

    <chemicalequationinput size="50"/>

    options: size -- width of the textbox.
    """

    template = "chemicalequationinput.html"
    tags = ["chemicalequationinput"]

    @classmethod
    def get_attributes(cls):
        """
        Can set size of text field.
        """
        return [
            Attribute("size", "20"),
        ]

    def _extra_context(self):
        """
        TODO (vshnayder): Get rid of this once we have a standard way of requiring js to be loaded.
        """
        return {
            "previewer": f"{self.capa_system.STATIC_URL}js/capa/chemical_equation_preview.js",
        }

    def handle_ajax(self, dispatch, data):
        """
        Since we only have chemcalc preview this input, check to see if it
        matches the corresponding dispatch and send it through if it does
        """
        if dispatch == "preview_chemcalc":
            return self.preview_chemcalc(data)
        return {}

    def preview_chemcalc(self, data):
        """
        Render an html preview of a chemical formula or equation.  get should
        contain a key 'formula' and value 'some formula string'.

        Returns a json dictionary::

            {
                'preview' : 'the-preview-html' or ''
                'error' : 'the-error' or ''
            }

        """

        _ = self.capa_system.i18n.gettext
        result = {"preview": "", "error": ""}
        try:
            formula = data["formula"]
        except KeyError:
            result["error"] = _("No formula specified.")
            return result

        try:
            result["preview"] = chemcalc.render_to_html(formula)
        except pyparsing.ParseException as err:
            result["error"] = _("Couldn't parse formula: {error_msg}").format(error_msg=err.msg)
        except Exception:  # pylint: disable=broad-exception-caught
            # this is unexpected, so log
            log.warning("Error while previewing chemical formula", exc_info=True)
            result["error"] = _("Error while rendering preview")

        return result


# -------------------------------------------------------------------------


@registry.register
class FormulaEquationInput(InputTypeBase):
    """
    An input type for entering formula equations.  Supports live preview.

    Example:

    <formulaequationinput size="50"/>

    options: size -- width of the textbox.
             trailing_text -- text to show after the input textbox when
                              rendered, same as textline (useful for units)
    """

    template = "formulaequationinput.html"
    tags = ["formulaequationinput"]

    @classmethod
    def get_attributes(cls):
        """
        Can set size of text field.
        """
        return [
            Attribute("size", "20"),
            Attribute("inline", False),
            Attribute("trailing_text", ""),
        ]

    def _extra_context(self):
        """
        TODO (vshnayder): Get rid of 'previewer' once we have a standard way of requiring js to be loaded.
        """
        # `reported_status` is basically `status`, except we say 'unanswered'

        return {
            "previewer": f"{self.capa_system.STATIC_URL}js/capa/src/formula_equation_preview.js",
        }

    def handle_ajax(self, dispatch, data):
        """
        Since we only have formcalc preview this input, check to see if it
        matches the corresponding dispatch and send it through if it does
        """
        if dispatch == "preview_formcalc":
            return self.preview_formcalc(data)
        return {}

    def preview_formcalc(self, get):
        """
        Render a preview of a formula or equation. `get` should
        contain a key 'formula' with a math expression.

        Returns a JSON dictionary::

            {
               'preview' : '<some latex>' or '',
               'error' : 'the-error' or '',
               'request_start' : <time sent with request>
            }
        """
        _ = self.capa_system.i18n.gettext
        result = {"preview": "", "error": ""}

        try:
            formula = get["formula"]
        except KeyError:
            result["error"] = _("No formula specified.")
            return result

        result["request_start"] = int(get.get("request_start", 0))

        # TODO add references to valid variables and functions
        # At some point, we might want to mark invalid variables as red
        # or something, and this is where we would need to pass those in.
        try:
            numeric_result = preview_numeric_input(formula)
            # Map results into the correct format
            result["preview"] = numeric_result["preview"]
            if numeric_result["error"]:
                result["error"] = numeric_result["error"]
            # if formula is invalid return formula
            if not numeric_result["is_valid"]:
                result["formula"] = formula
        except pyparsing.ParseException:
            result["error"] = _("Sorry, couldn't parse formula")
            result["formula"] = formula
        except Exception:  # pylint: disable=broad-exception-caught
            log.warning("Error while previewing formula", exc_info=True)
            result["error"] = _("Error while rendering preview")
            return result

        return result


def preview_numeric_input(formula):
    """
    Handles numeric validations, validates that the formula provided is a valid formula.
    """
    result = {"preview": "", "is_valid": True, "error": ""}
    try:
        result["preview"] = latex_preview(formula)
    except pyparsing.ParseException:
        result["error"] = "Sorry, couldn't parse formula"
        result["is_valid"] = False
        return result
    except Exception:  # pylint: disable=broad-exception-caught
        log.warning("Error while previewing formula", exc_info=True)
        result["error"] = "Error while rendering preview"
        result["is_valid"] = False
    return result


# -----------------------------------------------------------------------------


@registry.register
class DragAndDropInput(InputTypeBase):
    """
    Input for drag and drop problems. Allows student to drag and drop images and
    labels to base image.
    """

    template = "drag_and_drop_input.html"
    tags = ["drag_and_drop_input"]

    def setup(self):

        def parse(tag, tag_type):
            """Parses <tag ... /> xml element to dictionary. Stores
            'draggable' and 'target' tags with attributes to dictionary and
            returns last.

            Args:
                tag: xml etree element <tag...> with attributes

                tag_type: 'draggable' or 'target'.

                If tag_type is 'draggable' : all attributes except id
                (name or label or icon or can_reuse) are optional

                If tag_type is 'target' all attributes (name, x, y, w, h)
                are required. (x, y) - coordinates of center of target,
                w, h - weight and height of target.

            Returns:
                Dictionary of vaues of attributes:
                dict{'name': smth, 'label': smth, 'icon': smth,
                'can_reuse': smth}.
            """
            tag_attrs = {}
            tag_attrs["draggable"] = {
                "id": Attribute._sentinel,  # pylint: disable=protected-access
                "label": "",
                "icon": "",
                "can_reuse": "",
            }

            tag_attrs["target"] = {
                "id": Attribute._sentinel,  # pylint: disable=protected-access
                "x": Attribute._sentinel,  # pylint: disable=protected-access
                "y": Attribute._sentinel,  # pylint: disable=protected-access
                "w": Attribute._sentinel,  # pylint: disable=protected-access
                "h": Attribute._sentinel,  # pylint: disable=protected-access
            }

            dic = {}

            for attr_name in tag_attrs[tag_type].keys():
                dic[attr_name] = Attribute(attr_name, default=tag_attrs[tag_type][attr_name]).parse_from_xml(tag)

            if tag_type == "draggable" and not self.no_labels:
                dic["label"] = dic["label"] or dic["id"]

            if tag_type == "draggable":
                dic["target_fields"] = [parse(target, "target") for target in tag.iterchildren("target")]

            return dic

        # add labels to images?:
        self.no_labels = Attribute("no_labels", default="False").parse_from_xml(self.xml)

        to_js = {}

        # image drag and drop onto
        to_js["base_image"] = Attribute("img").parse_from_xml(self.xml)

        # outline places on image where to drag adn drop
        to_js["target_outline"] = Attribute("target_outline", default="False").parse_from_xml(self.xml)
        # one draggable per target?
        to_js["one_per_target"] = Attribute("one_per_target", default="True").parse_from_xml(self.xml)
        # list of draggables
        to_js["draggables"] = [parse(draggable, "draggable") for draggable in self.xml.iterchildren("draggable")]
        # list of targets
        to_js["targets"] = [parse(target, "target") for target in self.xml.iterchildren("target")]

        # custom background color for labels:
        label_bg_color = Attribute("label_bg_color", default=None).parse_from_xml(self.xml)
        if label_bg_color:
            to_js["label_bg_color"] = label_bg_color

        self.loaded_attributes["drag_and_drop_json"] = json.dumps(to_js)
        self.to_render.add("drag_and_drop_json")


# -------------------------------------------------------------------------


@registry.register
class DesignProtein2dInput(InputTypeBase):
    """
    An input type for design of a protein in 2D. Integrates with the Protex java applet.

    Example:

    <designprotein2d width="800" hight="500" target_shape="E;NE;NW;W;SW;E;none" />
    """

    template = "designprotein2dinput.html"
    tags = ["designprotein2dinput"]

    @classmethod
    def get_attributes(cls):
        """
        Note: width, hight, and target_shape are required.
        """
        return [Attribute("width"), Attribute("height"), Attribute("target_shape")]

    def _extra_context(self):
        context = {
            "applet_loader": f"{self.capa_system.STATIC_URL}js/capa/design-protein-2d.js",
        }

        return context


# -----------------------------------------------------------------------------


@registry.register
class EditAGeneInput(InputTypeBase):
    """
    An input type for editing a gene.
    Integrates with the genex GWT application.

    Example:

    <editagene genex_dna_sequence="CGAT" genex_problem_number="1"/>
    """

    template = "editageneinput.html"
    tags = ["editageneinput"]

    @classmethod
    def get_attributes(cls):
        """
        Note: width, height, and dna_sequencee are required.
        """
        return [Attribute("genex_dna_sequence"), Attribute("genex_problem_number")]

    def _extra_context(self):
        context = {
            "applet_loader": f"{self.capa_system.STATIC_URL}js/capa/edit-a-gene.js",
        }

        return context


# ---------------------------------------------------------------------


@registry.register
class AnnotationInput(InputTypeBase):
    """Input type for annotations.

    Students can enter some notes or other text (currently ungraded), and
    then choose from a set of tags/options, which are graded.

    Example::

        <annotationinput>
            <title>Annotation Exercise</title>
            <text>
                They are the ones who, at the public assembly, had put savage derangement [ate] into my thinking
                [phrenes] |89 on that day when I myself deprived Achilles of his honorific portion [geras]
            </text>
            <comment>Agamemnon says that ate or 'derangement' was the cause of his actions:
            why could Zeus say the same thing?</comment>
            <comment_prompt>Type a commentary below:</comment_prompt>
            <tag_prompt>Select one tag:</tag_prompt>
            <options>
                <option choice="correct">ate - both a cause and an effect</option>
                <option choice="incorrect">ate - a cause</option>
                <option choice="partially-correct">ate - an effect</option>
            </options>
        </annotationinput>

    # TODO: allow ordering to be randomized
    """

    template = "annotationinput.html"
    tags = ["annotationinput"]

    def setup(self):
        xml = self.xml

        self.debug = False  # set to True to display extra debug info with input
        self.return_to_annotation = True  # return only works in conjunction with annotatable xmodule

        self.title = xml.findtext("./title", "Annotation Exercise")
        self.text = xml.findtext("./text")
        self.comment = xml.findtext("./comment")
        self.comment_prompt = xml.findtext("./comment_prompt", "Type a commentary below:")
        self.tag_prompt = xml.findtext("./tag_prompt", "Select one tag:")
        self.options = self._find_options()

        # Need to provide a value that JSON can parse if there is no
        # student-supplied value yet.
        if self.value == "":
            self.value = "null"

        self._validate_options()

    def _find_options(self):
        """Returns an array of dicts where each dict represents an option."""
        elements = self.xml.findall("./options/option")
        return [
            {"id": index, "description": option.text, "choice": option.get("choice")}
            for (index, option) in enumerate(elements)
        ]

    def _validate_options(self):
        """Raises a ValueError if the choice attribute is missing or invalid."""
        valid_choices = ("correct", "partially-correct", "incorrect")
        for option in self.options:
            choice = option["choice"]
            if choice is None:
                raise ValueError("Missing required choice attribute.")
            if choice not in valid_choices:
                raise ValueError(f"Invalid choice attribute: {choice}. Must be one of: {', '.join(valid_choices)}")

    def _unpack(self, json_value):
        """Unpacks the json input state into a dict."""
        d = json.loads(json_value)
        if not isinstance(d, dict):
            d = {}

        comment_value = d.get("comment", "")
        if not isinstance(comment_value, str):
            comment_value = ""

        options_value = d.get("options", [])
        if not isinstance(options_value, list):
            options_value = []

        return {
            "options_value": options_value,
            "has_options_value": len(options_value) > 0,  # for convenience
            "comment_value": comment_value,
        }

    def _extra_context(self):
        extra_context = {
            "title": self.title,
            "text": self.text,
            "comment": self.comment,
            "comment_prompt": self.comment_prompt,
            "tag_prompt": self.tag_prompt,
            "options": self.options,
            "return_to_annotation": self.return_to_annotation,
            "debug": self.debug,
        }

        extra_context.update(self._unpack(self.value))

        return extra_context


@registry.register
class ChoiceTextGroup(InputTypeBase):
    r"""Groups of radiobutton/checkboxes with text inputs.

    Examples:
    **RadioButton Problem**::

        <problem>
            <startouttext/>
            A person rolls a standard die 100 times and records the results.
            On the first roll they received a "1". Given this information
            select the correct choice and fill in numbers to make it accurate.
            <endouttext/>
            <choicetextresponse>
                <radiotextgroup>
                    <choice correct="false">The lowest number rolled was:
                        <decoy_input/> and the highest number rolled was:
                        <decoy_input/> .
                    </choice>
                    <choice correct="true">The lowest number rolled was <numtolerance_input answer="1"/>
                        and there is not enough information to determine the highest number rolled.
                    </choice>
                    <choice correct="false">There is not enough information to determine the lowest
                        number rolled, and the highest number rolled was:
                        <decoy_input/> .
                    </choice>
                </radiotextgroup>
            </choicetextresponse>
        </problem>

    **Checkbox Problem**::

        <problem>
            <startouttext/>
            A person randomly selects 100 times, with replacement, from the list of numbers \(\sqrt{2}\), 2, 3, 4, 5, 6
            and records the results. The first number they pick is \(\sqrt{2}\). Given this information
            select the correct choices and fill in numbers to make them accurate.
            <endouttext/>
            <choicetextresponse>
                <checkboxtextgroup>
                    <choice correct="true">
                        The lowest number selected was <numtolerance_input answer="1.4142" tolerance="0.01"/>
                    </choice>
                    <choice correct="false">
                        The highest number selected was <decoy_input/> .
                    </choice>
                    <choice correct="true">There is not enough information given to determine the highest number
                        which was selected.
                    </choice>
                    <choice correct="false">There is not enough information given to determine the lowest number
                        selected.
                    </choice>
                </checkboxtextgroup>
            </choicetextresponse>
        </problem>

    .. note::
        In the preceding examples the <decoy_input/> is used to generate a textinput html element
        in the problem's display. Since it is inside of an incorrect choice, no answer given
        for it will be correct, and thus specifying an answer for it is not needed.
    """

    template = "choicetext.html"
    tags = ["radiotextgroup", "checkboxtextgroup"]

    def setup(self):
        """
        Performs setup for the initial rendering of the problem.
        `self.html_input_type` determines whether this problem is displayed
        with radiobuttons or checkboxes

        If the initial value of `self.value` is '' change it to {} so that
        the template has an empty dictionary to work with.

        sets the value of self.choices to be equal to the return value of
        `self.extract_choices`
        """
        self.text_input_values = {}
        if self.tag == "radiotextgroup":
            self.html_input_type = "radio"
        elif self.tag == "checkboxtextgroup":
            self.html_input_type = "checkbox"
        else:
            _ = self.capa_system.i18n.gettext
            msg = _("{input_type}: unexpected tag {tag_name}").format(input_type="ChoiceTextGroup", tag_name=self.tag)
            raise Exception(msg)  # pylint: disable=broad-exception-raised

        if self.value == "":
            # Make `value` an empty dictionary, if it currently has an empty
            # value. This is necessary because the template expects a
            # dictionary.
            self.value = {}
        self.choices = self.extract_choices(self.xml, self.capa_system.i18n)

    @classmethod
    def get_attributes(cls):
        """
        Returns a list of `Attribute` for this problem type
        """

        # Make '_' a no-op so we can scrape strings. Using function instead of
        #  `django.utils.translation.ugettext_noop` because Django cannot be imported in this file
        def _(text):
            return text

        return [
            Attribute("show_correctness", "always"),
            Attribute("submitted_message", _("Answer received.")),
        ]

    def _extra_context(self):
        """
        Returns a dictionary of extra content necessary for rendering this InputType.

        `input_type` is either 'radio' or 'checkbox' indicating whether the choices for
        this problem will have radiobuttons or checkboxes.
        """
        return {"input_type": self.html_input_type, "choices": self.choices}

    @staticmethod
    def extract_choices(element, i18n):
        """Extracts choices from the xml for this problem type.

        If we have xml that is as follows(choice names will have been assigned
        by now)::

            <radiotextgroup>
                <choice correct="true" name="1_2_1_choiceinput_0bc">
                    The number
                    <numtolerance_input name="1_2_1_choiceinput0_numtolerance_input_0" answer="5"/>
                    Is the mean of the list.
                </choice>
                <choice correct="false" name="1_2_1_choiceinput_1bc">
                    False demonstration choice
                </choice>
            </radiotextgroup>

        Choices are used for rendering the problem properly. The function
        will set up choices as follows::

            choices = [
                ("1_2_1_choiceinput_0bc",
                    [{'type': 'text', 'contents': "The number", 'tail_text': '',
                      'value': ''
                     },
                     {'type': 'textinput',
                      'contents': "1_2_1_choiceinput0_numtolerance_input_0",
                      'tail_text': 'Is the mean of the list',
                      'value': ''
                     }
                    ]
                ),
                ("1_2_1_choiceinput_1bc",
                    [{'type': 'text', 'contents': "False demonstration choice",
                      'tail_text': '',
                      'value': ''
                     }
                    ]
                )
            ]
        """

        _ = i18n.gettext
        choices = []

        for choice in element:
            if choice.tag != "choice":
                msg = Text("[capa.inputtypes.extract_choices] {0}").format(
                    # Translators: a "tag" is an XML element, such as "<b>" in HTML
                    Text(_("Expected a {expected_tag} tag; got {given_tag} instead")).format(
                        # xss-lint: disable=python-wrap-html
                        expected_tag="<choice>",
                        given_tag=choice.tag,
                    )
                )
                raise Exception(msg)  # pylint: disable=broad-exception-raised

            components = []
            choice_text = ""
            if choice.text is not None:
                choice_text += choice.text
            # Initialize our dict for the next content
            adder = {"type": "text", "contents": choice_text, "tail_text": "", "value": ""}
            components.append(adder)

            for elt in choice:
                # for elements in the choice e.g. <text> <numtolerance_input>
                adder = {"type": "text", "contents": "", "tail_text": "", "value": ""}
                tag_type = elt.tag
                # If the current `elt` is a <numtolerance_input> set the
                # `adder`type to 'numtolerance_input', and 'contents' to
                # the `elt`'s name.
                # Treat decoy_inputs and numtolerance_inputs the same in order
                # to prevent students from reading the Html and figuring out
                # which inputs are valid
                if tag_type in ("numtolerance_input", "decoy_input"):
                    # We set this to textinput, so that we get a textinput html
                    # element.
                    adder["type"] = "textinput"
                    adder["contents"] = elt.get("name")
                else:
                    adder["contents"] = elt.text

                # Add any tail text("is the mean" in the example)
                adder["tail_text"] = elt.tail if elt.tail else ""
                components.append(adder)

            # Add the tuple for the current choice to the list of choices
            choices.append((choice.get("name"), components))
        return choices



# pylint: disable=too-many-lines
#
# File:   capa/capa_problem.py
#
# Nomenclature:
#
# A capa Problem is a collection of text and capa Response questions.
# Each Response may have one or more Input entry fields.
# The capa problem may include a solution.
#
"""
Main module which shows problems (of "capa" type).

This is used by capa_block.
"""

import logging
import os.path
import re
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from typing import Optional
from xml.sax.saxutils import unescape
from zoneinfo import ZoneInfo

from django.conf import settings
from lxml import etree

from xblocks_contrib.problem.capa import customrender, inputtypes, responsetypes, xqueue_interface
from xblocks_contrib.problem.capa.correctmap import CorrectMap
from xblocks_contrib.problem.capa.safe_exec import safe_exec
from xblocks_contrib.problem.capa.util import (
    contextualize_text,
    convert_files_to_filenames,
    get_course_id_from_capa_block,
)
from xblocks_contrib.problem.markup import HTML, Text
from xblocks_contrib.problem.stringify import stringify_children
from xblocks_contrib.problem.xmlparser import XML

# extra things displayed after "show answers" is pressed
solution_tags = ["solution"]

# fully accessible capa input types
ACCESSIBLE_CAPA_INPUT_TYPES = [
    "checkboxgroup",
    "radiogroup",
    "choicegroup",
    "optioninput",
    "textline",
    "formulaequationinput",
    "textbox",
]

# these get captured as student responses
response_properties = ["codeparam", "responseparam", "answer", "openendedparam"]

# special problem tags which should be turned into innocuous HTML
html_transforms = {
    "problem": {"tag": "div"},
    "text": {"tag": "span"},
    "math": {"tag": "span"},
}

# These should be removed from HTML output, including all subelements
html_problem_semantics = [
    "additional_answer",
    "codeparam",
    "responseparam",
    "answer",
    "script",
    "hintgroup",
    "openendedparam",
    "openendedrubric",
]

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# main class for this module


class LoncapaSystem:
    """
    An encapsulation of resources needed from the outside.

    These interfaces are collected here so that a caller of LoncapaProblem
    can provide these resources however make sense for their environment, and
    this code can remain independent.

    Attributes:
        i18n: an object implementing the `gettext.Translations` interface so
            that we can use `.ugettext` to localize strings.

    See :class:`ModuleStoreRuntime` for documentation of other attributes.

    """

    def __init__(  # pylint: disable=too-many-positional-arguments,too-many-arguments
        self,
        ajax_url,
        anonymous_student_id,
        cache,
        can_execute_unsafe_code,
        get_python_lib_zip,
        DEBUG,  # pylint: disable=invalid-name
        i18n,
        render_template,
        resources_fs,
        seed,  # Why do we do this if we have self.seed?
        xqueue,
        matlab_api_key=None,
    ):
        self.ajax_url = ajax_url
        self.anonymous_student_id = anonymous_student_id
        self.cache = cache
        self.can_execute_unsafe_code = can_execute_unsafe_code
        self.get_python_lib_zip = get_python_lib_zip
        self.DEBUG = DEBUG  # pylint: disable=invalid-name
        self.i18n = i18n
        self.render_template = render_template
        self.resources_fs = resources_fs
        self.seed = seed  # Why do we do this if we have self.seed?
        self.STATIC_URL = settings.STATIC_URL  # pylint: disable=invalid-name
        self.xqueue = xqueue
        self.matlab_api_key = matlab_api_key


class LoncapaProblem:
    """
    Main class for capa Problems.
    """

    def __init__(  # pylint: disable=too-many-positional-arguments,too-many-arguments
        self,
        problem_text,
        id,  # pylint: disable=redefined-builtin
        capa_system,
        capa_block,
        state=None,
        seed=None,
        minimal_init=False,
        extract_tree=True,
    ):
        """
        Initializes capa Problem.

        Arguments:

            problem_text (string): xml defining the problem.
            id (string): identifier for this problem, often a filename (no spaces).
            capa_system (LoncapaSystem): LoncapaSystem instance which provides OS,
                rendering, user context, and other resources.
            capa_block: instance needed to access runtime/logging
            state (dict): containing the following keys:
                - `seed` (int) random number generator seed
                - `student_answers` (dict) maps input id to the stored answer for that input
                - 'has_saved_answers' (Boolean) True if the answer has been saved since last submit.
                - `correct_map` (CorrectMap) a map of each input to their 'correctness'
                - `done` (bool) indicates whether or not this problem is considered done
                - `input_state` (dict) maps input_id to a dictionary that holds the state for that input
            seed (int): random number generator seed.
            minimal_init (bool): whether to skip pre-processing student answers
            extract_tree (bool): whether to parse the problem XML and store the HTML

        """

        # Initialize class variables from state
        self.do_reset()
        self.problem_id = id
        self.capa_system = capa_system
        self.capa_block = capa_block

        state = state or {}

        # Set seed according to the following priority:
        #       1. Contained in problem's state
        #       2. Passed into capa_problem via constructor
        self.seed = state.get("seed", seed)
        assert self.seed is not None, "Seed must be provided for LoncapaProblem."

        self.student_answers = state.get("student_answers", {})
        self.has_saved_answers = state.get("has_saved_answers", False)
        if "correct_map" in state:
            self.correct_map.set_dict(state["correct_map"])
            self.correct_map_history = []
            for cmap in state.get("correct_map_history", []):
                correct_map = CorrectMap()
                correct_map.set_dict(cmap)
                self.correct_map_history.append(correct_map)

        self.done = state.get("done", False)
        self.input_state = state.get("input_state", {})

        # Convert startouttext and endouttext to proper <text></text>
        problem_text = re.sub(r"startouttext\s*/", "text", problem_text)
        problem_text = re.sub(r"endouttext\s*/", "/text", problem_text)
        self.problem_text = problem_text

        # parse problem XML file into an element tree
        if isinstance(problem_text, str):
            # etree chokes on Unicode XML with an encoding declaration
            problem_text = problem_text.encode("utf-8")
        self.tree = XML(problem_text)

        try:
            self.make_xml_compatible(self.tree)
        except Exception:
            capa_block = self.capa_block
            log.exception(
                "CAPAProblemError: %s, id:%s, data: %s", capa_block.display_name, self.problem_id, capa_block.data
            )
            raise

        # handle any <include file="foo"> tags
        self._process_includes()

        # construct script processor context (eg for customresponse problems)
        if minimal_init:
            self.context = {}
        else:
            self.context = self._extract_context(self.tree)

        # Pre-parse the XML tree: modifies it to add ID's and perform some in-place
        # transformations.  This also creates the dict (self.responders) of Response
        # instances for each question in the problem. The dict has keys = xml subtree of
        # Response, values = Response instance
        self.problem_data = self._preprocess_problem(self.tree, minimal_init)

        if not minimal_init:
            if not self.student_answers:  # True when student_answers is an empty dict
                self.set_initial_display()

            # dictionary of InputType objects associated with this problem
            #   input_id string -> InputType object
            self.inputs = {}

            # Run response late_transforms last (see MultipleChoiceResponse)
            # Sort the responses to be in *_1 *_2 ... order.
            responses = list(self.responders.values())
            responses = sorted(responses, key=lambda resp: int(resp.id[resp.id.rindex("_") + 1:]))
            for response in responses:
                if hasattr(response, "late_transforms"):
                    response.late_transforms(self)

            if extract_tree:
                self.extracted_tree = self._extract_html(self.tree)

    def make_xml_compatible(self, tree):
        """
        Adjust tree xml in-place for compatibility before creating
        a problem from it.

        The idea here is to provide a central point for XML translation,
        for example, supporting an old XML format. At present, there just two translations.

        1. <additional_answer> compatibility translation:

           old::

               <additional_answer>ANSWER</additional_answer>

           convert to

           new::

               <additional_answer answer="ANSWER">OPTIONAL-HINT</addional_answer>

        2. <optioninput> compatibility translation:

           optioninput works like this internally::

               <optioninput options="('yellow','blue','green')" correct="blue" />

           With extended hints there is a new <option> tag, like this::

               <option correct="True">blue <optionhint>sky color</optionhint></option>

           This translation takes in the new format and synthesizes the old
           ``option=`` attribute so all downstream logic works unchanged with
           the new <option> tag format.
        """

        def is_optioninput_valid(optioninput):
            """
            Verifies if a given optioninput xml is valid or not.

            A given optioninput(Dropdown) problem is invalid if it has more than one correct answer.

            Argument:
                optioninput: dropdown specification tree
            Returns:
                boolean: signifying if the optioninput is valid or not.
            """
            correct_options = [option.get("correct").upper() == "TRUE" for option in optioninput.findall("./option")]
            return correct_options.count(True) in (0, 1)

        additionals = tree.xpath("//stringresponse/additional_answer")
        for additional in additionals:
            answer = additional.get("answer")
            text = additional.text
            if not answer and text:  # trigger of old->new conversion
                additional.set("answer", text)
                additional.text = ""
        for optioninput in tree.xpath("//optioninput"):
            if not is_optioninput_valid(optioninput):
                raise responsetypes.LoncapaProblemError("Dropdown questions can only have one correct answer.")
            correct_option = None
            child_options = []
            for option_element in optioninput.findall("./option"):
                text = option_element.text
                text = text or ""
                option_name = text.strip()
                if option_element.get("correct").upper() == "TRUE":
                    correct_option = option_name
                child_options.append("'" + option_name + "'")

            if len(child_options) > 0:
                options_string = "(" + ",".join(child_options) + ")"
                optioninput.attrib.update({"options": options_string})
                if correct_option:
                    optioninput.attrib.update({"correct": correct_option})

    def do_reset(self):
        """
        Reset internal state to unfinished, with no answers
        """
        self.student_answers = {}
        self.student_answers_history = []
        self.has_saved_answers = False
        self.correct_map = CorrectMap()
        self.correct_map_history = []
        self.done = False

    def set_initial_display(self):
        """
        Set the student's answers to the responders' initial displays, if specified.
        """
        initial_answers = {}
        for responder in self.responders.values():
            if hasattr(responder, "get_initial_display"):
                initial_answers.update(responder.get_initial_display())

        self.student_answers = initial_answers

    def __str__(self):
        return f"LoncapaProblem ({self.problem_id})"

    def get_state(self):
        """
        Stored per-user session data neeeded to:
            1) Recreate the problem
            2) Populate any student answers.
        """

        return {
            "seed": self.seed,
            "student_answers": self.student_answers,
            "has_saved_answers": self.has_saved_answers,
            "correct_map": self.correct_map.get_dict(),
            "correct_map_history": [cmap.get_dict() for cmap in self.correct_map_history],
            "input_state": self.input_state,
            "done": self.done,
        }

    def get_max_score(self):
        """
        Return the maximum score for this problem.
        """
        maxscore = 0
        for responder in self.responders.values():
            maxscore += responder.get_max_score()
        return maxscore

    def calculate_score(self, correct_map=None):
        """
        Compute score for this problem.  The score is the number of points awarded.
        Returns a dictionary::

            {
                'score': integer, from 0 to get_max_score(),
                'total': get_max_score()
            }

        Takes an optional correctness map for use in the rescore workflow.
        """
        if correct_map is None:
            correct_map = self.correct_map
        correct = 0
        for key in correct_map:
            try:
                correct += correct_map.get_npoints(key)
            except Exception:
                log.error("key=%s, correct_map = %s", key, correct_map)
                raise

        return {"score": correct, "total": self.get_max_score()}

    def update_score(self, score_msg, queuekey):
        """
        Deliver grading response (e.g. from async code checking) to
            the specific ResponseType that requested grading

        Returns an updated CorrectMap
        """
        cmap = CorrectMap()
        cmap.update(self.correct_map)
        for responder in self.responders.values():
            if hasattr(responder, "update_score"):
                # Each LoncapaResponse will update its specific entries in cmap
                #   cmap is passed by reference
                responder.update_score(score_msg, cmap, queuekey)
        self.correct_map.set_dict(cmap.get_dict())
        return cmap

    def ungraded_response(self, xqueue_msg, queuekey):
        """
        Handle any responses from the xqueue that do not contain grades
        Will try to pass the queue message to all inputtypes that can handle ungraded responses

        Does not return any value
        """
        # check against each inputtype
        for the_input in self.inputs.values():
            # if the input type has an ungraded function, pass in the values
            if hasattr(the_input, "ungraded_response"):
                the_input.ungraded_response(xqueue_msg, queuekey)

    def is_queued(self):
        """
        Returns True if any part of the problem has been submitted to an external queue
        (e.g. for grading.)
        """
        return any(self.correct_map.is_queued(answer_id) for answer_id in self.correct_map)

    def get_recentmost_queuetime(self):
        """
        Returns a DateTime object that represents the timestamp of the most recent
        queueing request, or None if not queued
        """
        if not self.is_queued():
            return None

        # Get a list of timestamps of all queueing requests, then convert it to a DateTime object
        queuetime_strs = [
            self.correct_map.get_queuetime_str(answer_id)
            for answer_id in self.correct_map
            if self.correct_map.is_queued(answer_id)
        ]
        queuetimes = [
            datetime.strptime(qt_str, xqueue_interface.DATEFORMAT).replace(tzinfo=ZoneInfo("UTC"))
            for qt_str in queuetime_strs
        ]

        return max(queuetimes)

    def grade_answers(self, answers):
        """
        Grade student responses.  Called by capa_block.submit_problem.

        `answers` is a dict of all the entries from request.POST, but with the first part
        of each key removed (the string before the first "_").

        Thus, for example, input_ID123 -> ID123, and input_fromjs_ID123 -> fromjs_ID123

        Calls the Response for each question in this problem, to do the actual grading.
        """

        # if answers include File objects, convert them to filenames.
        self.student_answers = convert_files_to_filenames(answers)
        new_cmap = self.get_grade_from_current_answers(answers)
        self.correct_map = new_cmap  # pylint: disable=attribute-defined-outside-init
        self.correct_map_history.append(deepcopy(new_cmap))
        return self.correct_map

    def supports_rescoring(self):
        """
        Checks that the current problem definition permits rescoring.

        More precisely, it checks that there are no response types in
        the current problem that are not fully supported (yet) for rescoring.

        This includes responsetypes for which the student's answer
        is not properly stored in state, i.e. file submissions.  At present,
        we have no way to know if an existing response was actually a real
        answer or merely the filename of a file submitted as an answer.

        It turns out that because rescoring is a background task, limiting
        it to responsetypes that don't support file submissions also means
        that the responsetypes are synchronous.  This is convenient as it
        permits rescoring to be complete when the rescoring call returns.
        """
        return all("filesubmission" not in responder.allowed_inputfields for responder in self.responders.values())

    def get_grade_from_current_answers(self, student_answers, correct_map: Optional[CorrectMap] = None):
        """
        Gets the grade for the currently-saved problem state, but does not save it
        to the block.

        For new student_answers being graded, `student_answers` is a dict of all the
        entries from request.POST, but with the first part of each key removed
        (the string before the first "_").  Thus, for example,
        input_ID123 -> ID123, and input_fromjs_ID123 -> fromjs_ID123.

        For rescoring, `student_answers` is None.

        Calls the Response for each question in this problem, to do the actual grading.

        When the grading method is enabled, this method is used for rescore. In this case,
        the `correct_map` and the `student_answers` passed as arguments will be used,
        corresponding to each pair in the fields that store the history (correct_map_history
        and student_answers_history). The correct map will always be updated, depending on
        the student answers. The student answers will always remain the same over time.
        """
        oldcmap = correct_map if correct_map is not None else self.correct_map

        # start new with empty CorrectMap
        newcmap = CorrectMap()
        # Call each responsetype instance to do actual grading
        for responder in self.responders.values():
            # File objects are passed only if responsetype explicitly allows
            # for file submissions.  But we have no way of knowing if
            # student_answers contains a proper answer or the filename of
            # an earlier submission, so for now skip these entirely.
            # TODO: figure out where to get file submissions when rescoring.
            if "filesubmission" in responder.allowed_inputfields and student_answers is None:
                _ = self.capa_system.i18n.gettext
                raise Exception(  # pylint: disable=broad-exception-raised
                    _("Cannot rescore problems with possible file submissions")
                )

            # use 'student_answers' only if it is provided, and if it might contain a file
            # submission that would not exist in the persisted "student_answers".
            # If grading method is enabled, we need to pass each student answers and the
            # correct map in the history fields.
            results = responder.evaluate_answers(student_answers, oldcmap)
            newcmap.update(results)

        return newcmap

    def get_question_answers(self):
        """
        Returns a dict of answer_ids to answer values. If we cannot generate
        an answer (this sometimes happens in customresponses), that answer_id is
        not included. Called by "show answers" button JSON request
        (see capa_block)
        """
        # dict of (id, correct_answer)
        answer_map = {}
        for response in self.responders:
            results = self.responder_answers[response]
            answer_map.update(results)

        # include solutions from <solution>...</solution> stanzas
        for entry in self.tree.xpath("//" + "|//".join(solution_tags)):
            answer = etree.tostring(entry).decode("utf-8")
            if answer:
                answer_map[entry.get("id")] = contextualize_text(answer, self.context)

        log.debug("answer_map = %s", answer_map)
        return answer_map

    def get_answer_ids(self):
        """
        Return the IDs of all the responses -- these are the keys used for
        the dicts returned by grade_answers and get_question_answers. (Though
        get_question_answers may only return a subset of these.
        """
        answer_ids = []
        for response in self.responders:
            results = self.responder_answers[response]
            answer_ids.append(list(results.keys()))
        return answer_ids

    def find_correct_answer_text(self, answer_id):
        """
        Returns the correct answer(s) for the provided answer_id as a single string.

        Arguments::
            answer_id (str): a string like "98e6a8e915904d5389821a94e48babcf_13_1"

        Returns:
            str: A string containing the answer or multiple answers separated by commas.
        """
        xml_elements = self.tree.xpath('//*[@id="' + answer_id + '"]')
        if not xml_elements:
            return None
        xml_element = xml_elements[0]
        answer_text = xml_element.xpath("@answer")
        if answer_text:
            return answer_id[0]
        if xml_element.tag == "optioninput":
            return xml_element.xpath("@correct")[0]
        return ", ".join(xml_element.xpath('*[@correct="true"]/text()'))

    def find_question_label(self, answer_id):
        """
        Obtain the most relevant question text for a particular answer.

        E.g. in a problem like "How much is 2+2?" "Two"/"Three"/"More than three",
        this function returns the "How much is 2+2?" text.

        It uses, in order:
        - the question prompt, if the question has one
        - the <p> or <label> element which precedes the choices (skipping descriptive elements)
        - a text like "Question 5" if no other name could be found

        Arguments::
            answer_id: a string like "98e6a8e915904d5389821a94e48babcf_13_1"

        Returns:
            a string with the question text
        """

        def generate_default_question_label():
            """
            To create question string like "Question 2" by adding "Question" and its position number.
            For instance 'd2e35c1d294b4ba0b3b1048615605d2a_2_1' contains 2,
            which is used in question number 1 (see example XML in comment above)
            There's no question 0 (question IDs start at 1, answer IDs at 2)
            """
            question_nr = int(answer_id.split("_")[-2]) - 1
            return _("Question {}").format(question_nr)

        _ = self.capa_system.i18n.gettext
        # Some questions define a prompt with this format:   >>This is a prompt<<
        try:
            prompt = self.problem_data[answer_id].get("label")
        except KeyError:
            prompt = None

        if prompt:
            question_text = prompt.striptags()
        else:
            # If no prompt, then we must look for something resembling a question ourselves
            #
            # We have a structure like:
            #
            # <p />
            # <optionresponse id="a0effb954cca4759994f1ac9e9434bf4_2">
            #   <optioninput id="a0effb954cca4759994f1ac9e9434bf4_3_1" />
            # <optionresponse>
            #
            # Starting from  answer (the optioninput in this example) we go up and backwards
            xml_elems = self.tree.xpath('//*[@id="' + answer_id + '"]')
            if len(xml_elems) != 1:
                return generate_default_question_label()

            xml_elem = xml_elems[0].getparent()

            # Get the element that probably contains the question text
            questiontext_elem = xml_elem.getprevious()

            # Go backwards looking for a <p> or <label>, but skip <description> because it doesn't
            # contain the question text.
            #
            # E.g if we have this:
            #   <p /> <description /> <optionresponse /> <optionresponse />
            #
            # then from the first optionresponse we'll end with the <p>.
            # If we start in the second optionresponse, we'll find another response in the way,
            # stop early, and instead of a question we'll report "Question 2".
            skip_elems = ["description"]
            label_elems = ["p", "label"]
            while questiontext_elem is not None and questiontext_elem.tag in skip_elems:
                questiontext_elem = questiontext_elem.getprevious()

            if questiontext_elem is not None and questiontext_elem.tag in label_elems:
                question_text = questiontext_elem.text
            else:
                question_text = generate_default_question_label()

        return question_text

    def find_answer_text(self, answer_id, current_answer):
        """
        Process a raw answer text to make it more meaningful.

        E.g. in a choice problem like "How much is 2+2?" "Two"/"Three"/"More than three",
        this function will transform "choice_1" (which is the internal response given by
        many capa methods) to the human version, e.g. "More than three".

        If the answers are multiple (e.g. because they're from a multiple choice problem),
        this will join them with a comma.

        If passed a normal string which is already the answer, it doesn't change it.

        TODO merge with response_a11y_data?

        Arguments:
            answer_id: a string like "98e6a8e915904d5389821a94e48babcf_13_1"
            current_answer: a data structure as found in `LoncapaProblem.student_answers`
                which represents the best response we have until now

        Returns:
            a string with the human version of the response
        """
        if isinstance(current_answer, list):
            # Multiple answers. This case happens e.g. in multiple choice problems
            answer_text = ", ".join(self.find_answer_text(answer_id, answer) for answer in current_answer)

        elif isinstance(current_answer, str) and current_answer.startswith("choice_"):
            # Many problem (e.g. checkbox) report "choice_0" "choice_1" etc.
            # Here we transform it
            elems = self.tree.xpath(f'//*[@id="{answer_id}"]//*[@name="{current_answer}"]')
            if len(elems) == 0:
                log.warning("Answer Text Missing for answer id: %s and choice number: %s", answer_id, current_answer)
                answer_text = "Answer Text Missing"
            elif len(elems) == 1:
                choicegroup = elems[0].getparent()
                input_cls = inputtypes.registry.get_class_for_tag(choicegroup.tag)
                choices_map = dict(input_cls.extract_choices(choicegroup, self.capa_system.i18n, text_only=True))
                answer_text = choices_map.get(current_answer, "Answer Text Missing")
            else:
                log.warning("Multiple answers found for answer id: %s and choice number: %s", answer_id, current_answer)
                answer_text = "Multiple answers found"

        elif isinstance(current_answer, str):
            # Already a string with the answer
            answer_text = current_answer

        else:
            raise NotImplementedError()

        return answer_text or "Answer Text Missing"

    def do_targeted_feedback(self, tree):  # pylint: disable=too-many-locals,too-many-branches
        """
        Implements targeted-feedback in-place on  <multiplechoiceresponse> --
        choice-level explanations shown to a student after submission.
        Does nothing if there is no targeted-feedback attribute.
        """
        _ = self.capa_system.i18n.gettext
        # Note that the modifications has been done, avoiding problems if called twice.
        if hasattr(self, "has_targeted"):
            return
        self.has_targeted = True  # pylint: disable=attribute-defined-outside-init

        for mult_choice_response in tree.xpath("//multiplechoiceresponse[@targeted-feedback]"):
            show_explanation = mult_choice_response.get("targeted-feedback") == "alwaysShowCorrectChoiceExplanation"

            # Grab the first choicegroup (there should only be one within each <multiplechoiceresponse> tag)
            choicegroup = mult_choice_response.xpath('./choicegroup[@type="MultipleChoice"]')[0]
            choices_list = list(choicegroup.iter("choice"))

            # Find the student answer key that matches our <choicegroup> id
            student_answer = self.student_answers.get(choicegroup.get("id"))
            expl_id_for_student_answer = None

            # Keep track of the explanation-id that corresponds to the student's answer
            # Also, keep track of the solution-id
            solution_id = None
            choice_correctness_for_student_answer = _("Incorrect")
            for choice in choices_list:
                if choice.get("name") == student_answer:
                    expl_id_for_student_answer = choice.get("explanation-id")
                    if choice.get("correct") == "true":
                        choice_correctness_for_student_answer = _("Correct")
                if choice.get("correct") == "true":
                    solution_id = choice.get("explanation-id")

            # Filter out targetedfeedback that doesn't correspond to the answer the student selected
            # Note: following-sibling will grab all following siblings, so we just want the first in the list
            targetedfeedbackset = mult_choice_response.xpath("./following-sibling::targetedfeedbackset")
            if len(targetedfeedbackset) != 0:
                targetedfeedbackset = targetedfeedbackset[0]
                targetedfeedbacks = targetedfeedbackset.xpath("./targetedfeedback")
                # find the legend by id in choicegroup.html for aria-describedby
                problem_legend_id = str(choicegroup.get("id")) + "-legend"
                for targetedfeedback in targetedfeedbacks:
                    screenreadertext = etree.Element("span")
                    targetedfeedback.insert(0, screenreadertext)
                    screenreadertext.set("class", "sr")
                    screenreadertext.text = choice_correctness_for_student_answer
                    targetedfeedback.set("role", "group")
                    targetedfeedback.set("aria-describedby", problem_legend_id)
                    # Don't show targeted feedback if the student hasn't answer the problem
                    # or if the target feedback doesn't match the student's (incorrect) answer
                    if not self.done or targetedfeedback.get("explanation-id") != expl_id_for_student_answer:
                        targetedfeedbackset.remove(targetedfeedback)

            # Do not displace the solution under these circumstances
            if not show_explanation or not self.done:
                continue

            # The next element should either be <solution> or <solutionset>
            next_element = targetedfeedbackset.getnext()
            parent_element = tree
            solution_element = None
            if next_element is not None and next_element.tag == "solution":
                solution_element = next_element
            elif next_element is not None and next_element.tag == "solutionset":
                solutions = next_element.xpath("./solution")
                for solution in solutions:
                    if solution.get("explanation-id") == solution_id:
                        parent_element = next_element
                        solution_element = solution

            # If could not find the solution element, then skip the remaining steps below
            if solution_element is None:
                continue

            # Change our correct-choice explanation from a "solution explanation" to within
            # the set of targeted feedback, which means the explanation will render on the page
            # without the student clicking "Show Answer" or seeing a checkmark next to the correct choice
            parent_element.remove(solution_element)

            # Add our solution instead to the targetedfeedbackset and change its tag name
            solution_element.tag = "targetedfeedback"

            targetedfeedbackset.append(solution_element)

    def get_html(self):
        """
        Main method called externally to get the HTML to be rendered for this capa Problem.
        """
        self.do_targeted_feedback(self.tree)
        html = contextualize_text(etree.tostring(self._extract_html(self.tree)).decode("utf-8"), self.context)
        return html

    def handle_input_ajax(self, data):
        """
        InputTypes can support specialized AJAX calls. Find the correct input and pass along the correct data

        Also, parse out the dispatch from the get so that it can be passed onto the input type nicely
        """

        # pull out the id
        input_id = data["input_id"]
        if self.inputs[input_id]:
            dispatch = data["dispatch"]
            return self.inputs[input_id].handle_ajax(dispatch, data)

        log.warning("Could not find matching input for id: %s", input_id)
        return {}

    # ======= Private Methods Below ========

    def _process_includes(self):
        """
        Handle any <include file="foo"> tags by reading in the specified file and inserting it
        into our XML tree.  Fail gracefully if debugging.
        """
        includes = self.tree.findall(".//include")
        for inc in includes:
            filename = inc.get("file")
            if filename is not None:
                try:
                    # open using LoncapaSystem OSFS filesystem
                    ifp = self.capa_system.resources_fs.open(filename)
                except Exception as err:  # pylint: disable=broad-exception-caught
                    log.warning("Error %s in problem xml include: %s", err, etree.tostring(inc, pretty_print=True))
                    log.warning("Cannot find file %s in %s", filename, self.capa_system.resources_fs)
                    # if debugging, don't fail - just log error
                    # TODO (vshnayder): need real error handling, display to users
                    if not self.capa_system.DEBUG:
                        raise
                    continue
                try:
                    # read in and convert to XML
                    incxml = etree.XML(ifp.read())
                except Exception as err:  # pylint: disable=broad-exception-caught
                    log.warning("Error %s in problem xml include: %s", err, etree.tostring(inc, pretty_print=True))
                    log.warning("Cannot parse XML in %s", (filename))
                    # if debugging, don't fail - just log error
                    # TODO (vshnayder): same as above
                    if not self.capa_system.DEBUG:
                        raise
                    continue

                # insert new XML into tree in place of include
                parent = inc.getparent()
                parent.insert(parent.index(inc), incxml)
                parent.remove(inc)
                log.debug("Included %s into %s", filename, self.problem_id)

    def _extract_system_path(self, script):
        """
        Extracts and normalizes additional paths for code execution.
        For now, there's a default path of data/course/code; this may be removed
        at some point.

        script : ?? (TODO)
        """

        default_path = ["code"]

        # Separate paths by :, like the system path.
        raw_path = script.get("system_path", "").split(":") + default_path

        # find additional comma-separated modules search path
        path = []

        for dir in raw_path:  # pylint: disable=redefined-builtin
            if not dir:
                continue

            # path is an absolute path or a path relative to the data dir
            dir = os.path.join(self.capa_system.resources_fs.root_path, dir)
            # Check that we are within the resources_fs tree.
            reldir = os.path.relpath(dir, self.capa_system.resources_fs.root_path)
            if ".." in reldir:
                log.warning("Ignoring Python directory outside of course: %r", dir)
                continue

            abs_dir = os.path.normpath(dir)
            path.append(abs_dir)

        return path

    def _extract_context(self, tree):
        """
        Extract content of <script>...</script> from the problem.xml file, and exec it in the
        context of this problem.  Provides ability to randomize problems, and also set
        variables for problem answer checking.

        Problem XML goes to Python execution context. Runs everything in script tags.
        """
        context = {}
        context["seed"] = self.seed
        context["anonymous_student_id"] = self.capa_system.anonymous_student_id
        all_code = ""

        python_path = []

        for script in tree.findall(".//script"):

            stype = script.get("type")
            if stype:
                if "javascript" in stype:
                    continue  # skip javascript
                if "perl" in stype:
                    continue  # skip perl
            # TODO: evaluate only python

            for d in self._extract_system_path(script):
                if d not in python_path and os.path.exists(d):
                    python_path.append(d)

            xmlesc = {"&apos;": "'", "&quot;": '"'}
            code = unescape(script.text, xmlesc)
            all_code += code

        extra_files = []
        if all_code:
            # An asset named python_lib.zip can be imported by Python code.
            zip_lib = self.capa_system.get_python_lib_zip()
            if zip_lib is not None:
                extra_files.append(("python_lib.zip", zip_lib))
                python_path.append("python_lib.zip")

            try:
                safe_exec(
                    all_code,
                    context,
                    random_seed=self.seed,
                    python_path=python_path,
                    extra_files=extra_files,
                    cache=self.capa_system.cache,
                    limit_overrides_context=get_course_id_from_capa_block(self.capa_block),
                    slug=self.problem_id,
                    unsafely=self.capa_system.can_execute_unsafe_code(),
                )
            except Exception as err:
                log.exception("Error while execing script code: %s", all_code)
                msg = Text(f"Error while executing script code: {err}")
                raise responsetypes.LoncapaProblemError(msg)

        # Store code source in context, along with the Python path needed to run it correctly.
        context["script_code"] = all_code
        context["python_path"] = python_path
        context["extra_files"] = extra_files or None
        return context

    def _extract_html(  # private
        self, problemtree
    ):  # pylint: disable=too-many-locals,too-many-branches,too-many-return-statements
        """
        Main (private) function which converts Problem XML tree to HTML.
        Calls itself recursively.

        Returns Element tree of XHTML representation of problemtree.
        Calls render_html of Response instances to render responses into XHTML.

        Used by get_html.
        """
        if not isinstance(problemtree.tag, str):
            # Comment and ProcessingInstruction nodes are not Elements,
            # and we're ok leaving those behind.
            # BTW: etree gives us no good way to distinguish these things
            # other than to examine .tag to see if it's a string. :(
            return None

        if problemtree.tag == "script" and problemtree.get("type") and "javascript" in problemtree.get("type"):
            # leave javascript intact.
            return deepcopy(problemtree)

        if problemtree.tag in html_problem_semantics:
            return None

        problemid = problemtree.get("id")  # my ID

        if problemtree.tag in inputtypes.registry.registered_tags():
            # If this is an inputtype subtree, let it render itself.
            response_data = self.problem_data[problemid]

            status = "unsubmitted"
            msg = ""
            hint = ""
            hintmode = None
            input_id = problemtree.get("id")
            answervariable = None
            if problemid in self.correct_map:
                pid = input_id

                # If we're withholding correctness, don't show adaptive hints either.
                # Note that regular, "demand" hints will be shown, if the course author has added them to the problem.
                if not self.capa_block.correctness_available():
                    status = "submitted"
                else:
                    # If the the problem has not been saved since the last submit set the status to the
                    # current correctness value and set the message as expected. Otherwise we do not want to
                    # display correctness because the answer may have changed since the problem was graded.
                    if not self.has_saved_answers:
                        status = self.correct_map.get_correctness(pid)
                        msg = self.correct_map.get_msg(pid)

                    hint = self.correct_map.get_hint(pid)
                    hintmode = self.correct_map.get_hintmode(pid)
                    answervariable = self.correct_map.get_property(pid, "answervariable")

            value = ""
            if self.student_answers and problemid in self.student_answers:
                value = self.student_answers[problemid]

            if input_id not in self.input_state:
                self.input_state[input_id] = {}

            # do the rendering
            state = {
                "value": value,
                "status": status,
                "id": input_id,
                "input_state": self.input_state[input_id],
                "answervariable": answervariable,
                "response_data": response_data,
                "has_saved_answers": self.has_saved_answers,
                "feedback": {
                    "message": msg,
                    "hint": hint,
                    "hintmode": hintmode,
                },
            }

            input_type_cls = inputtypes.registry.get_class_for_tag(problemtree.tag)
            # save the input type so that we can make ajax calls on it if we need to
            self.inputs[input_id] = input_type_cls(self.capa_system, problemtree, state)
            return self.inputs[input_id].get_html()

        # let each Response render itself
        if problemtree in self.responders:
            overall_msg = self.correct_map.get_overall_message()
            return self.responders[problemtree].render_html(self._extract_html, response_msg=overall_msg)

        # let each custom renderer render itself:
        if problemtree.tag in customrender.registry.registered_tags():
            renderer_class = customrender.registry.get_class_for_tag(problemtree.tag)
            renderer = renderer_class(self.capa_system, problemtree)
            return renderer.get_html()

        # otherwise, render children recursively, and copy over attributes
        tree = etree.Element(problemtree.tag)
        for item in problemtree:
            item_xhtml = self._extract_html(item)
            if item_xhtml is not None:
                tree.append(item_xhtml)

        if tree.tag in html_transforms:
            tree.tag = html_transforms[problemtree.tag]["tag"]
        else:
            # copy attributes over if not innocufying
            for key, value in problemtree.items():
                tree.set(key, value)

        tree.text = problemtree.text
        tree.tail = problemtree.tail

        return tree

    def _preprocess_problem(self, tree, minimal_init):  # private
        """
        Assign IDs to all the responses
        Assign sub-IDs to all entries (textline, schematic, etc.)
        Annoted correctness and value
        In-place transformation

        Also create capa Response instances for each responsetype and save as self.responders

        Obtain all responder answers and save as self.responder_answers dict (key = response)
        """
        response_id = 1
        problem_data = {}
        self.responders = {}
        for response in tree.xpath("//" + "|//".join(responsetypes.registry.registered_tags())):
            responsetype_id = self.problem_id + "_" + str(response_id)
            # create and save ID for this response
            response.set("id", responsetype_id)
            response_id += 1

            answer_id = 1
            input_tags = inputtypes.registry.registered_tags()
            inputfields = tree.xpath(
                "|".join(["//" + response.tag + "[@id=$id]//" + x for x in input_tags]), id=responsetype_id
            )

            # assign one answer_id for each input type
            for entry in inputfields:
                entry.attrib["response_id"] = str(response_id)
                entry.attrib["answer_id"] = str(answer_id)
                entry.attrib["id"] = f"{self.problem_id}_{response_id}_{answer_id}"
                answer_id = answer_id + 1

            self.response_a11y_data(response, inputfields, responsetype_id, problem_data)

            # instantiate capa Response
            responsetype_cls = responsetypes.registry.get_class_for_tag(response.tag)
            responder = responsetype_cls(
                response, inputfields, self.context, self.capa_system, self.capa_block, minimal_init
            )
            # save in list in self
            self.responders[response] = responder

        if not minimal_init:
            # get responder answers (do this only once, since there may be a performance cost,
            # eg with externalresponse)
            self.responder_answers = {}
            for response, responder in self.responders.items():
                try:
                    self.responder_answers[response] = responder.get_answers()
                except Exception:
                    log.debug("responder %s failed to properly return get_answers()", responder)  # FIXME
                    raise

            # <solution>...</solution> may not be associated with any specific response; give
            # IDs for those separately
            # TODO: We should make the namespaces consistent and unique (e.g. %s_problem_%i).
            solution_id = 1
            for solution in tree.findall(".//solution"):
                solution.attrib["id"] = f"{self.problem_id}_solution_{solution_id}"
                solution_id += 1

        return problem_data

    def response_a11y_data(  # pylint: disable=too-many-locals,too-many-branches
        self, response, inputfields, responsetype_id, problem_data
    ):
        """
        Construct data to be used for a11y.

        Arguments:
            response (object): xml response object
            inputfields (list): list of inputfields in a responsetype
            responsetype_id (str): responsetype id
            problem_data (dict): dict to be filled with response data
        """
        # if there are no inputtypes then don't do anything
        if not inputfields:
            return

        element_to_be_deleted = None
        label = ""

        if len(inputfields) > 1:
            response.set("multiple_inputtypes", "true")
            group_label_tag = response.find("label")
            group_description_tags = response.findall("description")
            group_label_tag_id = f"multiinput-group-label-{responsetype_id}"
            group_label_tag_text = ""
            if group_label_tag is not None:
                group_label_tag.tag = "p"
                group_label_tag.set("id", group_label_tag_id)
                group_label_tag.set("class", "multi-inputs-group-label")
                group_label_tag_text = stringify_children(group_label_tag)
                response.set("multiinput-group-label-id", group_label_tag_id)

            group_description_ids = []
            for index, group_description_tag in enumerate(group_description_tags):
                group_description_tag_id = f"multiinput-group-description-{responsetype_id}-{index}"
                group_description_tag.tag = "p"
                group_description_tag.set("id", group_description_tag_id)
                group_description_tag.set("class", "multi-inputs-group-description question-description")
                group_description_ids.append(group_description_tag_id)

            if group_description_ids:
                response.set("multiinput-group_description_ids", " ".join(group_description_ids))

            for inputfield in inputfields:
                problem_data[inputfield.get("id")] = {
                    "group_label": group_label_tag_text,
                    "label": HTML(inputfield.attrib.get("label", "")),
                    "descriptions": {},
                }
        else:
            # Extract label value from <label> tag or label attribute from inside the responsetype
            responsetype_label_tag = response.find("label")
            if responsetype_label_tag is not None:
                label = stringify_children(responsetype_label_tag)
                # store <label> tag containing question text to delete
                # it later otherwise question will be rendered twice
                element_to_be_deleted = responsetype_label_tag
            elif "label" in inputfields[0].attrib:
                # in this case we have old problems with label attribute and p tag having question in it
                # we will pick the first sibling of responsetype if its a p tag and match the text with
                # the label attribute text. if they are equal then we will use this text as question.
                # Get first <p> tag before responsetype, this <p> may contains the question text.
                p_tag = response.xpath("preceding-sibling::*[1][self::p]")

                if p_tag and p_tag[0].text == inputfields[0].attrib["label"]:
                    label = stringify_children(p_tag[0])
                    element_to_be_deleted = p_tag[0]
            else:
                # In this case the problems don't have tag or label attribute inside the responsetype
                # so we will get the first preceding label tag w.r.t to this responsetype.
                # This will take care of those multi-question problems that are not using --- in their markdown.
                label_tag = response.xpath("preceding-sibling::*[1][self::label]")
                if label_tag:
                    label = stringify_children(label_tag[0])
                    element_to_be_deleted = label_tag[0]

            # delete label or p element only if inputtype is fully accessible
            if inputfields[0].tag in ACCESSIBLE_CAPA_INPUT_TYPES and element_to_be_deleted is not None:
                element_to_be_deleted.getparent().remove(element_to_be_deleted)

            # Extract descriptions and set unique id on each description tag
            description_tags = response.findall("description")
            description_id = 1
            descriptions = OrderedDict()
            for description in description_tags:
                descriptions[f"description_{responsetype_id}_{description_id}"] = HTML(stringify_children(description))
                response.remove(description)
                description_id += 1

            problem_data[inputfields[0].get("id")] = {
                "label": HTML(label.strip()) if label else "",
                "descriptions": descriptions,
            }
